import gradio as gr
import numpy as np
import os
from qim3d.tools import internal_tools
from qim3d.io import DataLoader
import plotly.graph_objects as go
from scipy import ndimage


class Interface:
    def __init__(self):
        self.show_header = False
        self.verbose = False
        self.title = "Isosurfaces for 3D visualization"

        self.plot_height = 768
        self.height = 1024
        self.width = 960

        # Data examples
        current_dir = os.path.dirname(os.path.abspath(__file__))
        examples_dir = ["..", "..", "img_examples"]
        examples = [
            "blobs_256x256x256.tif",
            "fly_150x256x256.tif",
            "cement_128x128x128.tif",
            "NT_10x200x100.tif",
            "NT_128x128x128.tif",
            "shell_225x128x128.tif",
            "bone_128x128x128.tif",
        ]
        self.img_examples = []
        for example in examples:
            self.img_examples.append(
                [os.path.join(current_dir, *examples_dir, example)]
            )

        # CSS path
        self.css_path = os.path.join(current_dir, "..", "css", "gradio.css")

    def clear(self):
        """Used to reset the plot with the clear button"""
        return None

    def load_data(self, filepath):

        # TODO: Add support for multiple files
        self.vol = DataLoader().load_tiff(filepath)

    def resize_vol(self):
        """Resizes the loaded volume to the display size"""

        # Get original size
        original_Z, original_Y, original_X = np.shape(self.vol)
        max_size = np.max([original_Z, original_Y, original_X])
        if self.verbose:
            print(f"\nOriginal volume: {original_Z, original_Y, original_X}")

        # Resize for display
        self.vol = ndimage.zoom(
            input=self.vol,
            zoom=self.display_size / max_size,
            order=0,
            prefilter=False,
        )

        self.display_size_z, self.display_size_y, self.display_size_x = np.shape(
            self.vol
        )
        if self.verbose:
            print(
                f"Resized volume: {self.display_size_z, self.display_size_y, self.display_size_x}"
            )

    def save_fig(self, fig, filename):
        # Write Plotly figure to disk
        fig.write_html(filename)

    def create_fig(self):
        # Load volume
        self.load_data(self.gradio_file.name)

        # Resize data for display size
        self.resize_vol()

        # Flip Z
        if self.flip_z:
            self.vol = np.flip(self.vol, axis=0)

        # Create 3D grid
        Z, Y, X = np.mgrid[
            0 : self.display_size_z, 0 : self.display_size_y, 0 : self.display_size_x
        ]

        if self.only_wireframe:
            surface_fill = 0.2
        else:
            surface_fill = 1.0

        fig = go.Figure(
            go.Volume(
                z=Z.flatten(),
                y=Y.flatten(),
                x=X.flatten(),
                value=self.vol.flatten(),
                isomin=self.min_value * np.max(self.vol),
                isomax=self.max_value * np.max(self.vol),
                cmin=np.min(self.vol),
                cmax=np.max(self.vol),
                opacity=self.opacity,
                opacityscale=self.opacityscale,
                surface_count=self.surface_count,
                colorscale=self.colormap,
                slices_z=dict(
                    show=self.show_z_slice,
                    locations=[int(self.display_size_z * self.slice_z_location)],
                ),
                slices_y=dict(
                    show=self.show_y_slice,
                    locations=[int(self.display_size_y * self.slice_y_location)],
                ),
                slices_x=dict(
                    show=self.show_x_slice,
                    locations=[int(self.display_size_x * self.slice_x_location)],
                ),
                surface=dict(fill=surface_fill),
                caps=dict(
                    x_show=self.show_caps,
                    y_show=self.show_caps,
                    z_show=self.show_caps,
                ),
                showscale=self.show_colorbar,
                colorbar=dict(
                    thickness=8, outlinecolor="#fff", len=0.5, orientation="h"
                ),
                reversescale=self.reversescale,
                hoverinfo="skip",
            )
        )

        fig.update_layout(
            scene_xaxis_showticklabels=self.show_ticks,
            scene_yaxis_showticklabels=self.show_ticks,
            scene_zaxis_showticklabels=self.show_ticks,
            scene_xaxis_visible=self.show_axis,
            scene_yaxis_visible=self.show_axis,
            scene_zaxis_visible=self.show_axis,
            scene_aspectmode="data",
            height=self.plot_height,
            hovermode=False,
            scene_camera_eye=dict(x=2.0, y=-2.0, z=1.5),
        )

        return fig

    def process(self, *args):
        # Get args passed by Gradio
        # TODO: solve this in an automated way
        # Could Gradio pass kwargs instead of args?
        self.gradio_file = args[0]
        self.display_size = args[1]
        self.opacity = args[2]
        self.opacityscale = args[3]
        self.only_wireframe = args[4]
        self.min_value = args[5]
        self.max_value = args[6]
        self.surface_count = args[7]
        self.colormap = args[8]
        self.show_colorbar = args[9]
        self.reversescale = args[10]
        self.flip_z = args[11]
        self.show_axis = args[12]
        self.show_ticks = args[13]
        self.show_caps = args[14]
        self.show_z_slice = args[15]
        self.slice_z_location = args[16]
        self.show_y_slice = args[17]
        self.slice_y_location = args[18]
        self.show_x_slice = args[19]
        self.slice_x_location = args[20]

        # Create output figure
        fig = self.create_fig()

        # Save it to disk
        self.save_fig(fig, "iso3d.html")

        return fig, "iso3d.html"

    def create_interface(self):
        # Create gradio app

        with gr.Blocks(css=self.css_path) as gradio_interface:
            if self.show_header:
                gr.Markdown(
                    """
                    # 3D Visualization (isosurfaces)
                    This tool uses Plotly Volume (https://plotly.com/python/3d-volume-plots/) to create iso surfaces from voxels based on their intensity levels.
                    To optimize performance when generating visualizations, set the number of voxels (_display resolution_) and isosurfaces (_total surfaces_) to lower levels. 
                    """
                )

            with gr.Row():
                # Input and parameters column
                with gr.Column(scale=1, min_width=320):
                    with gr.Tab("Input"):
                        # File loader
                        gradio_file = gr.File(
                            show_label=False, elem_classes="file-input h-128"
                        )
                    with gr.Tab("Examples"):
                        gr.Examples(examples=self.img_examples, inputs=gradio_file)

                    # Run button
                    with gr.Row():
                        with gr.Column(scale=3, min_width=64):
                            btn_run = gr.Button(
                                value="Run 3D visualization", elem_classes="btn btn-run"
                            )
                        with gr.Column(scale=1, min_width=64):
                            btn_clear = gr.Button(
                                value="Clear", elem_classes="btn btn-clear"
                            )

                    with gr.Tab("Display"):
                        # Display options

                        display_size = gr.Slider(
                            32,
                            128,
                            step=4,
                            label="Display resolution",
                            info="Number of voxels for the largest dimension",
                            value=64,
                            elem_classes="",
                        )
                        surface_count = gr.Slider(
                            2, 16, step=1, label="Total iso-surfaces", value=6
                        )

                        show_caps = gr.Checkbox(value=False, label="Show surface caps")

                        with gr.Row():
                            opacityscale = gr.Dropdown(
                                choices=["uniform", "extremes", "min", "max"],
                                value="uniform",
                                label="Opacity scale",
                                info="Handles opacity acording to voxel value",
                            )

                            opacity = gr.Slider(
                                0.0, 1.0, step=0.1, label="Max opacity", value=0.4
                            )
                        with gr.Row():
                            min_value = gr.Slider(
                                0.0, 1.0, step=0.05, label="Min value", value=0.1
                            )
                            max_value = gr.Slider(
                                0.0, 1.0, step=0.05, label="Max value", value=1
                            )

                    with gr.Tab("Slices"):
                        show_z_slice = gr.Checkbox(value=False, label="Show Z slice")
                        slice_z_location = gr.Slider(
                            0.0, 1.0, step=0.05, value=0.5, label="Position"
                        )

                        show_y_slice = gr.Checkbox(value=False, label="Show Y slice")
                        slice_y_location = gr.Slider(
                            0.0, 1.0, step=0.05, value=0.5, label="Position"
                        )

                        show_x_slice = gr.Checkbox(value=False, label="Show X slice")
                        slice_x_location = gr.Slider(
                            0.0, 1.0, step=0.05, value=0.5, label="Position"
                        )

                    with gr.Tab("Misc"):
                        with gr.Row():
                            colormap = gr.Dropdown(
                                choices=[
                                    "Blackbody",
                                    "Bluered",
                                    "Blues",
                                    "Cividis",
                                    "Earth",
                                    "Electric",
                                    "Greens",
                                    "Greys",
                                    "Hot",
                                    "Jet",
                                    "Magma",
                                    "Picnic",
                                    "Portland",
                                    "Rainbow",
                                    "RdBu",
                                    "Reds",
                                    "Viridis",
                                    "YlGnBu",
                                    "YlOrRd",
                                ],
                                value="Magma",
                                label="Colormap",
                            )

                            show_colorbar = gr.Checkbox(
                                value=False, label="Show color scale"
                            )
                            reversescale = gr.Checkbox(
                                value=False, label="Reverse color scale"
                            )
                        flip_z = gr.Checkbox(value=True, label="Flip Z axis")
                        show_axis = gr.Checkbox(value=True, label="Show axis")
                        show_ticks = gr.Checkbox(value=False, label="Show ticks")
                        only_wireframe = gr.Checkbox(
                            value=False, label="Only wireframe"
                        )

                    # Inputs for gradio
                    inputs = [
                        gradio_file,
                        display_size,
                        opacity,
                        opacityscale,
                        only_wireframe,
                        min_value,
                        max_value,
                        surface_count,
                        colormap,
                        show_colorbar,
                        reversescale,
                        flip_z,
                        show_axis,
                        show_ticks,
                        show_caps,
                        show_z_slice,
                        slice_z_location,
                        show_y_slice,
                        slice_y_location,
                        show_x_slice,
                        slice_x_location,
                    ]

                # Output display column
                with gr.Column(scale=4):
                    volvizplot = gr.Plot(show_label=False)

                    plot_download = gr.File(
                        interactive=False,
                        label="Download interactive plot",
                        show_label=True,
                        visible=False,
                        elem_classes="w-256",
                    )

                outputs = [volvizplot, plot_download]

            # Session for user data
            session = gr.State([])

            # Listeners

            # Clear button
            for gr_obj in outputs:
                btn_clear.click(fn=self.clear, inputs=None, outputs=gr_obj)
            # Run button
            # fmt: off
            btn_run.click(
                fn=self.process, inputs=inputs, outputs=outputs).success(
                fn=self.make_visible, inputs=None, outputs=plot_download)
            # fmt: on

        return gradio_interface

    def make_visible(self):
        return gr.update(visible=True)

    def launch(self, **kwargs):
        # Show header
        if self.show_header:
            internal_tools.gradio_header(self.title, self.port)

        # Create gradio interface
        interface = self.create_interface()

        # Set gradio verbose level
        if self.verbose:
            quiet = False
        else:
            quiet = True

        interface.launch(
            quiet=quiet,
            height=self.height,
            width=self.width,
            **kwargs,
        )


if __name__ == "__main__":
    app = Interface()
    app.show_header = True
    app.launch(server_name="0.0.0.0", show_error=True, default_port=True)



import gradio as gr
import numpy as np
import os
from qim3d.tools import internal_tools
from qim3d.io import DataLoader
import tifffile
import plotly.express as px
from scipy import ndimage
import outputformat as ouf
import plotly.graph_objects as go
import localthickness as lt
import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt


class Interface:
    def __init__(self):
        self.show_header = False
        self.verbose = False
        self.title = "Local thickness"
        self.plot_height = 768
        self.height = 1024
        self.width = 960

        # Data examples
        current_dir = os.path.dirname(os.path.abspath(__file__))
        examples_dir = ["..", "..", "img_examples"]
        examples = [
            "blobs_256x256x256.tif",
            "cement_128x128x128.tif",
            "bone_128x128x128.tif",
            "NT_10x200x100.tif",
        ]
        self.img_examples = []
        for example in examples:
            self.img_examples.append(
                [os.path.join(current_dir, *examples_dir, example)]
            )

        # CSS path
        self.css_path = os.path.join(current_dir, "..", "css", "gradio.css")

    def clear(self):
        """Used to reset the plot with the clear button"""
        return None

    def make_visible(self):
        return gr.update(visible=True)

    def start_session(self, *args):
        session = Session()
        session.verbose = self.verbose
        session.interface = "gradio"

        # Get the args passed by gradio
        session.data = args[0]
        session.lt_scale = args[1]
        session.threshold = args[2]
        session.dark_objects = args[3]
        session.flip_z = args[4]
        session.nbins = args[5]
        session.display_size_input = args[6]
        session.surface_count_input = args[7]
        session.display_size_output = args[8]
        session.surface_count_output = args[9]
        session.reversescale = args[10]
        session.show_caps = args[11]

        return session

    def launch(self, **kwargs):
        # Show header
        if self.show_header:
            internal_tools.gradio_header(self.title, self.port)

        # Create gradio interfaces
        interface = self.create_interface()

        # Set gradio verbose level
        if self.verbose:
            quiet = False
        else:
            quiet = True

        interface.launch(
            quiet=quiet,
            height=self.height,
            width=self.width,
            **kwargs,
        )

    def create_interface(self):
        with gr.Blocks(css=self.css_path) as gradio_interface:
            gr.Markdown(
                "# 3D Local thickness \n Interface for _Fast local thickness in 3D and 2D_ (https://github.com/vedranaa/local-thickness)"
            )

            with gr.Row():
                with gr.Column(scale=1, min_width=320):
                    with gr.Tab("Input"):
                        data = gr.File(
                            show_label=False, elem_classes="file-input h-128"
                        )
                    with gr.Tab("Examples"):
                        gr.Examples(examples=self.img_examples, inputs=data)

                    # Run button
                    with gr.Row():
                        with gr.Column(scale=3, min_width=64):
                            btn = gr.Button(
                                "Run local thickness", elem_classes="btn btn-run"
                            )
                        with gr.Column(scale=1, min_width=64):
                            btn_clear = gr.Button("Clear", elem_classes="btn btn-clear")

                    with gr.Tab("Parameters"):
                        gr.Markdown(
                            "It is possible to scale down the image before processing. Lower values will make the algorithm run faster, but decreases the accuracy of results."
                        )
                        lt_scale = gr.Slider(
                            0.1, 1.0, label="Scale", value=0.5, step=0.1
                        )

                        with gr.Row():
                            threshold = gr.Slider(
                                0.0,
                                1.0,
                                value=0.5,
                                step=0.05,
                                label="Threshold",
                                info="Local thickness uses a binary image, so a threshold value is needed.",
                            )

                        dark_objects = gr.Checkbox(value=False, label="Dark objects")

                    with gr.Tab("Display"):
                        with gr.Row():
                            gr.Markdown("Input display")

                            display_size_input = gr.Slider(
                                16,
                                64,
                                step=4,
                                label="Display resolution",
                                info="Number of voxels for the largest dimension",
                                value=32,
                            )

                            surface_count_input = gr.Slider(
                                2, 16, step=1, label="Total iso-surfaces", value=4
                            )
                        with gr.Row():
                            gr.Markdown("Output display")

                            display_size_output = gr.Slider(
                                32,
                                128,
                                step=4,
                                label="Display resolution",
                                info="Number of voxels for the largest dimension",
                                value=64,
                            )

                            surface_count_output = gr.Slider(
                                2, 16, step=1, label="Total iso-surfaces", value=12
                            )

                        reversescale = gr.Checkbox(
                            value=False, label="Reverse color scale"
                        )

                        show_caps = gr.Checkbox(value=True, label="Show surface caps")

                        flip_z = gr.Checkbox(value=True, label="Flip Z axis")

                        gr.Markdown("Thickness histogram options")
                        nbins = gr.Slider(
                            5, 50, value=25, step=1, label="Histogram bins"
                        )

                    inputs = [
                        data,
                        lt_scale,
                        threshold,
                        dark_objects,
                        flip_z,
                        nbins,
                        display_size_input,
                        surface_count_input,
                        display_size_output,
                        surface_count_output,
                        reversescale,
                        show_caps,
                    ]

                with gr.Column(scale=4):
                    with gr.Row():
                        with gr.Column(min_width=256):
                            input_vol = gr.Plot(
                                show_label=True, label="Original volume", visible=True
                            )

                        with gr.Column(min_width=256):
                            binary_vol = gr.Plot(
                                show_label=True, label="Binary volume", visible=True
                            )

                    with gr.Row():
                        with gr.Column(min_width=256):
                            output_vol = gr.Plot(
                                show_label=True,
                                label="Local thickness volume",
                                visible=True,
                            )
                        with gr.Column(min_width=256):
                            histogram = gr.Plot(
                                show_label=True,
                                label="Thickness histogram",
                                visible=True,
                            )
                    with gr.Row():
                        with gr.Column():
                            lt_output = gr.File(
                                interactive=False,
                                show_label=True,
                                label="Output file",
                                visible=False,
                                elem_classes="w-320",
                            )

            # Pipelines
            pipeline = Pipeline()
            pipeline.verbose = self.verbose

            # Session
            session = gr.State([])

            # Ouput gradio objects
            outputs = [input_vol, output_vol, binary_vol, histogram, lt_output]

            # Clear button
            for gr_obj in outputs:
                btn_clear.click(fn=self.clear, inputs=None, outputs=gr_obj)

            # Run button
            # fmt: off
            btn.click(
                fn=self.start_session, inputs=inputs, outputs=session).success(
                fn=pipeline.process_input, inputs=session, outputs=session).success(
                fn=pipeline.prepare_volume, inputs=session,outputs=session).success(
                fn=pipeline.input_viz, inputs=session, outputs=input_vol).success(
                fn=pipeline.make_binary, inputs=session, outputs=session).success(
                fn=pipeline.binary_viz, inputs=session, outputs=binary_vol).success(
                fn=pipeline.compute_localthickness, inputs=session, outputs=session).success(
                fn=pipeline.prepare_output_for_display, inputs=session, outputs=session).success(
                fn=pipeline.output_viz, inputs=session, outputs=output_vol).success(
                fn=pipeline.thickness_histogram, inputs=session, outputs=histogram).success(
                fn=pipeline.save_lt, inputs=session, outputs=lt_output).success(
                fn=self.make_visible, inputs=None, outputs=lt_output)

            # fmt: on

        return gradio_interface


class Session:
    def __init__(self):
        self.interface = None
        self.verbose = None
        self.show_ticks = False
        self.show_axis = True

        # Args from gradio
        self.data = None
        self.vol = None
        self.lt_scale = None
        self.threshold = 0.5
        self.dark_objects = False
        self.flip_z = True
        self.nbins = 25
        self.display_size_input = 32
        self.surface_count_input = 4
        self.display_size_output = 64
        self.surface_count_output = 12
        self.reversescale = False
        self.show_caps = True

        # From pipeline
        self.vol_input_display = None
        self.input_display_size_z = None
        self.input_display_size_y = None
        self.input_display_size_x = None
        self.Zgrid = None
        self.Ygrid = None
        self.Xgrid = None
        self.vol_binary = None
        self.vol_binary_display = None
        self.vol_thickness = None
        self.output_display_size_z = None
        self.output_display_size_y = None
        self.output_display_size_x = None
        self.Zgrid_output = None
        self.Ygrid_output = None
        self.Xgrid_output = None

    def get_vol_info(self):
        self.original_shape = np.shape(self.vol)
        self.original_Z = self.original_shape[0]
        self.original_Y = self.original_shape[1]
        self.original_X = self.original_shape[2]
        self.max_size = np.max(self.original_shape)

        if self.verbose:
            print(f"Original volume shape:{self.original_shape}")
            print(f"Original Z: {self.original_Z}")
            print(f"Original Y: {self.original_Y}")
            print(f"Original X: {self.original_X}")
            print(f"Max size: {self.max_size}")


class Pipeline:
    def process_input(self, session):
        # Load volume
        session.vol = DataLoader().load(session.data.name)

        if session.dark_objects:
            session.vol = np.invert(session.vol)

        if session.flip_z:
            session.vol = np.flip(session.vol, axis=0)

        return session

    def prepare_volume(self, session):
        # Get volume shape
        session.get_vol_info()

        # Resize for display
        session.vol_input_display = ndimage.zoom(
            input=session.vol,
            zoom=(session.display_size_input / session.max_size),
            prefilter=False,
            order=0,
        )

        display_shape = np.shape(session.vol_input_display)
        session.input_display_size_z = display_shape[0]
        session.input_display_size_y = display_shape[1]
        session.input_display_size_x = display_shape[2]

        # Create 3D grid
        session.Zgrid, session.Ygrid, session.Xgrid = np.mgrid[
            0 : session.input_display_size_z,
            0 : session.input_display_size_y,
            0 : session.input_display_size_x,
        ]

        return session

    def input_viz(self, session):
        # Generate input visualization
        data = go.Volume(
            z=session.Zgrid.flatten(),
            y=session.Ygrid.flatten(),
            x=session.Xgrid.flatten(),
            value=session.vol_input_display.flatten(),
            opacity=0.3,
            isomin=0.05 * np.max(session.vol_input_display),
            isomax=1.0 * np.max(session.vol_input_display),
            cmin=np.min(session.vol_input_display),
            cmax=np.max(session.vol_input_display),
            opacityscale="uniform",
            surface_count=session.surface_count_input,
            caps=dict(
                x_show=session.show_caps,
                y_show=session.show_caps,
                z_show=session.show_caps,
            ),
            showscale=False,
            reversescale=session.reversescale,
        )

        fig = go.Figure(data)
        fig.update_layout(
            scene_aspectmode="data",
            scene_xaxis_showticklabels=session.show_ticks,
            scene_yaxis_showticklabels=session.show_ticks,
            scene_zaxis_showticklabels=session.show_ticks,
            scene_xaxis_visible=session.show_axis,
            scene_yaxis_visible=session.show_axis,
            scene_zaxis_visible=session.show_axis,
            hovermode=False,
            scene_camera_eye=dict(x=1.5, y=-1.5, z=1.2),
        )

        return fig

    def make_binary(self, session):
        # Make a binary volume
        # Nothing fancy, but we could add new features here
        session.vol_binary = session.vol > (session.threshold * np.max(session.vol))

        session.vol_binary_display = ndimage.zoom(
            input=session.vol_binary * 255,
            zoom=(session.display_size_input / session.max_size),
            prefilter=False,
            order=0,
        )

        return session

    def binary_viz(self, session):
        # Generate input visualization
        data = go.Volume(
            z=session.Zgrid.flatten(),
            y=session.Ygrid.flatten(),
            x=session.Xgrid.flatten(),
            value=session.vol_binary_display.flatten(),
            opacity=0.3,
            isomin=0.99 * np.max(session.vol_binary_display),
            isomax=1.0 * np.max(session.vol_binary_display),
            cmin=np.min(session.vol_binary_display),
            cmax=np.max(session.vol_binary_display),
            opacityscale="max",
            surface_count=2,
            caps=dict(
                x_show=session.show_caps,
                y_show=session.show_caps,
                z_show=session.show_caps,
            ),
            showscale=False,
            reversescale=session.reversescale,
            colorscale="Greys",
        )

        fig = go.Figure(data)
        fig.update_layout(
            scene_aspectmode="data",
            scene_xaxis_showticklabels=session.show_ticks,
            scene_yaxis_showticklabels=session.show_ticks,
            scene_zaxis_showticklabels=session.show_ticks,
            scene_xaxis_visible=session.show_axis,
            scene_yaxis_visible=session.show_axis,
            scene_zaxis_visible=session.show_axis,
            hovermode=False,
            scene_camera_eye=dict(x=1.5, y=-1.5, z=1.2),
        )

        return fig

    def compute_localthickness(self, session):
        session.vol_thickness = lt.local_thickness(session.vol_binary, session.lt_scale)

        return session

    def prepare_output_for_display(self, session):
        # Display Local thickness
        session.vol_output_display = ndimage.zoom(
            input=session.vol_thickness,
            zoom=(session.display_size_output / session.max_size),
            prefilter=False,
            order=0,
        )

        output_display_shape = np.shape(session.vol_output_display)
        session.output_display_size_z = output_display_shape[0]
        session.output_display_size_y = output_display_shape[1]
        session.output_display_size_x = output_display_shape[2]

        session.Zgrid_output, session.Ygrid_output, session.Xgrid_output = np.mgrid[
            0 : session.output_display_size_z,
            0 : session.output_display_size_y,
            0 : session.output_display_size_x,
        ]

        return session

    def output_viz(self, session):
        # Generate input visualization
        data = go.Volume(
            z=session.Zgrid_output.flatten(),
            y=session.Ygrid_output.flatten(),
            x=session.Xgrid_output.flatten(),
            value=session.vol_output_display.flatten(),
            opacity=0.3,
            isomin=0.05 * np.max(session.vol_output_display),
            isomax=1.0 * np.max(session.vol_output_display),
            cmin=np.min(session.vol_output_display),
            cmax=np.max(session.vol_output_display),
            opacityscale="uniform",
            surface_count=session.surface_count_input,
            caps=dict(
                x_show=session.show_caps,
                y_show=session.show_caps,
                z_show=session.show_caps,
            ),
            showscale=False,
            reversescale=session.reversescale,
        )

        fig = go.Figure(data)
        fig.update_layout(
            scene_aspectmode="data",
            scene_xaxis_showticklabels=session.show_ticks,
            scene_yaxis_showticklabels=session.show_ticks,
            scene_zaxis_showticklabels=session.show_ticks,
            scene_xaxis_visible=session.show_axis,
            scene_yaxis_visible=session.show_axis,
            scene_zaxis_visible=session.show_axis,
            hovermode=False,
            scene_camera_eye=dict(x=1.5, y=-1.5, z=1.2),
        )

        return fig

    def thickness_histogram(self, session):
        # Ignore zero thickness
        non_zero_values = session.vol_thickness[session.vol_thickness > 0]

        # Calculate histogram
        vol_hist, bin_edges = np.histogram(non_zero_values, session.nbins)

        fig, ax = plt.subplots(figsize=(6, 4))

        ax.bar(
            bin_edges[:-1], vol_hist, width=np.diff(bin_edges), ec="white", align="edge"
        )

        # Adjustments
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["left"].set_visible(True)
        ax.spines["bottom"].set_visible(True)
        ax.set_yscale("log")

        return fig

    def save_lt(self, session):
        filename = "localthickness.tif"
        # Save output image in a temp space
        tifffile.imwrite(filename, session.vol_thickness)

        return filename


def gradio_fn(
    gradio_file,
    lt_scale,
    threshold,
    dark_objects,
    flip_z,
    nbins,
    display_size_input,
    surface_count_input,
    display_size_output,
    surface_count_output,
    reversescale,
    show_caps,
    show_ticks=False,
    show_axis=True,
):
    # Some cleanup
    vol_input = None
    vol_input_display = None
    vol_output = None
    vol_output_display = None
    data = None

    return fig_input, fig_output, fig_binary, fig_hist, "localthickness.tif"


if __name__ == "__main__":
    app = Interface()
    app.show_header = True
    app.launch(server_name="0.0.0.0", show_error=True)



import tifffile
import h5py
import os
import difflib


class DataLoader:
    def __init__(self, **kwargs):
        self.verbose = False

        # Virtual stack is False by default
        self.virtual_stack = kwargs.get("virtual_stack", False)

    def load_tiff(self, path):
        if self.virtual_stack:
            vol = tifffile.memmap(path)
        else:
            vol = tifffile.imread(path)

        return vol

    def load_h5(self, path):
        with h5py.File(path, "r") as f:
            vol = f["data"][:]
        return vol

    def load(self, path):
        # Load a single file
        if os.path.isfile(path):
            if path.endswith(".tif") or path.endswith(".tiff"):
                return self.load_tiff(path)
            elif path.endswith(".h5"):
                return self.load_h5(path)
            else:
                raise ValueError("Unsupported file format")

        # Load a directory
        elif os.path.isdir(path):
            raise NotImplementedError("Loading from directory is not implemented yet")

        # Fails
        else:
            # Find the closest matching path to warn the user
            parent_dir = os.path.dirname(path)
            parent_files = os.listdir(parent_dir)
            valid_paths = [os.path.join(parent_dir, file) for file in parent_files]
            similar_paths = difflib.get_close_matches(path, valid_paths)
            if similar_paths:
                suggestion = similar_paths[0]  # Get the closest match
                message = f"Invalid path.\nDid you mean '{suggestion}'?"
                raise ValueError(message)
            else:
                raise ValueError("Invalid path")


def load(path, **kwargs):
    return DataLoader(**kwargs).load(path)



from setuptools import setup, find_packages
import os


# Read the contents of your README file
with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="qim3d",
    version="0.1.0",
    author="Felipe Delestro",
    author_email="fima@dtu.dk",
    description="QIM tools and user interfaces",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://lab.compute.dtu.dk/QIM/qim",
    packages=find_packages(),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Development Status :: 1 - Planning",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Image Processing",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: User Interfaces",
    ],
    python_requires=">=3.6",
    install_requires=['gradio>=3.31.0',
                      'h5py>=3.8.0',
                      'localthickness>=0.1.2',
                      'matplotlib>=3.7.1',
                      'numpy>=1.24.3',
                      'outputformat>=0.1.3',
                      'plotly>=5.14.1',
                      'scipy>=1.10.1',
                      'setuptools>=59.6.0',
                      'tifffile>=2023.4.12']
)



import gradio as gr
import numpy as np
import os
from qim3d.tools import internal_tools
from qim3d.io import DataLoader
import tifffile
import outputformat as ouf
import datetime
import matplotlib
matplotlib.use("Agg")
import matplotlib.pyplot as plt


class Interface:
    def __init__(self):
        self.show_header = False
        self.verbose = False
        self.title = "Data Explorer"

        self.height = 1024
        self.width = 900

        # CSS path
        current_dir = os.path.dirname(os.path.abspath(__file__))
        self.css_path = os.path.join(current_dir, "..", "css", "gradio.css")

    def clear(self):
        """Used to reset outputs with the clear button"""
        return None

    def create_interface(self):
        with gr.Blocks(css=self.css_path) as gradio_interface:
            gr.Markdown("# Data Explorer \n Quick insights from large datasets")

            with gr.Row():
                data_path = gr.Textbox(
                    value="/home/fima/Downloads/MarineGatropod_1.tif",
                    max_lines=1,
                    label="Path to the 3D volume",
                )
            with gr.Row(elem_classes=" w-128"):
                btn_run = gr.Button(value="Load & Run", elem_classes="btn btn-run")

            # Outputs
            with gr.Row():
                gr.Markdown("## Data overview")

            with gr.Row():
                data_summary = gr.Text(
                    label=None, show_label=False, elem_classes="monospace-box"
                )
                with gr.Column():
                    zslice_plot = gr.Plot(label="Z slice", elem_classes="rounded")
                    zpos = gr.Slider(
                        minimum=0, maximum=1, value=0.5, step=0.01, label="Z position"
                    )
                with gr.Column():
                    yslice_plot = gr.Plot(label="Y slice", elem_classes="rounded")

                    ypos = gr.Slider(
                        minimum=0, maximum=1, value=0.5, step=0.01, label="Y position"
                    )

                with gr.Column():
                    xslice_plot = gr.Plot(label="X slice", elem_classes="rounded")

                    xpos = gr.Slider(
                        minimum=0, maximum=1, value=0.5, step=0.01, label="X position"
                    )
            with gr.Row(elem_classes="h-32"):
                gr.Markdown()

            with gr.Row(elem_classes="h-480"):
                max_projection_plot = gr.Plot(
                    label="Z max projection", elem_classes="rounded"
                )
                min_projection_plot = gr.Plot(
                    label="Z min projection", elem_classes="rounded"
                )

                hist_plot = gr.Plot(label="Volume intensity histogram")

            pipeline = Pipeline()
            pipeline.verbose = self.verbose
            session = gr.State([])

            with gr.Row():
                gr.Markdown("## Local thickness")
            with gr.Row():
                gr.Plot()
                gr.Plot()
                gr.Plot()

            with gr.Row():
                gr.Markdown("## Structure tensor")
            with gr.Row():
                gr.Plot()
                gr.Plot()
                gr.Plot()

            ### Gradio objects lists

            # Inputs
            inputs = [zpos, ypos, xpos]
            # Outputs
            outputs = [
                data_summary,
                zslice_plot,
                yslice_plot,
                xslice_plot,
                max_projection_plot,
                min_projection_plot,
            ]

            projection_outputs = [session, max_projection_plot, min_projection_plot]

            ### Listeners
            # Clear button
            # for gr_obj in outputs:
            #     btn_clear.click(fn=self.clear, inputs=[], outputs=gr_obj)

            # Run button
            # fmt: off
            btn_run.click(
                fn=self.start_session, inputs=inputs, outputs=session).success(
                fn=pipeline.process_input, inputs=[session, data_path], outputs=session).success(
                fn=pipeline.show_summary_str, inputs=session, outputs=data_summary).success(
                fn=pipeline.create_zslice_fig, inputs=session, outputs=zslice_plot).success(
                fn=pipeline.create_yslice_fig, inputs=session, outputs=yslice_plot).success(
                fn=pipeline.create_xslice_fig, inputs=session, outputs=xslice_plot).success(
                fn=pipeline.create_projections_figs, inputs=session, outputs=projection_outputs).success(
                fn=pipeline.show_summary_str, inputs=session, outputs=data_summary).success(
                fn=pipeline.plot_vol_histogram, inputs=session, outputs=hist_plot)
                

            zpos.release(
                fn=self.update_zpos, inputs=[session, zpos], outputs=[session, zslice_plot]).success(
                fn=pipeline.create_zslice_fig, inputs=session, outputs=zslice_plot,show_progress=False)
            ypos.release(
                fn=self.update_ypos, inputs=[session, ypos], outputs=[session, yslice_plot]).success(
                fn=pipeline.create_yslice_fig, inputs=session, outputs=yslice_plot,show_progress=False)
            
            xpos.release(
                fn=self.update_xpos, inputs=[session, xpos], outputs=[session, xslice_plot]).success(
                fn=pipeline.create_xslice_fig, inputs=session, outputs=xslice_plot,show_progress=False)

            # fmt: on

        return gradio_interface

    def start_session(self, *args):
        # Starts a new session dictionary
        session = Session()
        session.interface = "gradio"
        session.zpos = args[0]
        session.ypos = args[1]
        session.xpos = args[2]

        return session

    def update_zpos(self, session, zpos):
        session.zpos = zpos
        session.zslice_from_zpos()

        return session, gr.update(label=f"Z slice: {session.zslice}")

    def update_ypos(self, session, ypos):
        session.ypos = ypos
        session.yslice_from_ypos()

        return session, gr.update(label=f"Y slice: {session.yslice}")

    def update_xpos(self, session, xpos):
        session.xpos = xpos
        session.xslice_from_xpos()

        return session, gr.update(label=f"X slice: {session.xslice}")

    def launch(self, **kwargs):
        # Show header
        if self.show_header:
            internal_tools.gradio_header(self.title, self.port)

        # Create gradio interfaces
        interface = self.create_interface()

        # Set gradio verbose level
        if self.verbose:
            quiet = False
        else:
            quiet = True


        interface.launch(
            quiet=quiet,
            height=self.height,
            width=self.width,
            **kwargs,
        )


class Session:
    def __init__(self):
        self.interface = None
        self.data_path = None
        self.vol = None
        self.zpos = 0.5
        self.ypos = 0.5
        self.xpos = 0.5

        # Volume info
        self.zsize = None
        self.ysize = None
        self.xsize = None
        self.data_type = None
        self.axes = None
        self.last_modified = None
        self.file_size = None
        self.min_percentile = None
        self.max_percentile = None
        self.min_value = None
        self.max_value = None
        self.intensity_sum = None
        self.mean_intensity = None

        # Histogram
        self.nbins = 32

    def get_data_info(self):
        # Open file
        tif = tifffile.TiffFile(self.data_path)
        first_slice = tif.pages[0]

        # Get info
        self.zsize = len(tif.pages)
        self.ysize, self.xsize = first_slice.shape
        self.data_type = first_slice.dtype
        self.axes = tif.series[0].axes
        self.last_modified = datetime.datetime.fromtimestamp(
            os.path.getmtime(self.data_path)
        ).strftime("%Y-%m-%d %H:%M")
        self.file_size = os.path.getsize(self.data_path)

        # Close file
        tif.close()

    def create_summary_dict(self):
        # Create dictionary
        self.summary_dict = {
            "Last modified": self.last_modified,
            "File size": internal_tools.sizeof(self.file_size),
            "Axes": self.axes,
            "Z-size": str(self.zsize),
            "Y-size": str(self.ysize),
            "X-size": str(self.xsize),
            "Data type": self.data_type,
            "Min value": self.min_value,
            "Mean value": self.mean_intensity,
            "Max value": self.max_value,
        }

    def summary_str(self):
        display_dict = {k: v for k, v in self.summary_dict.items() if v is not None}
        return ouf.showdict(display_dict, return_str=True, title="Data summary")

    def zslice_from_zpos(self):
        self.zslice = int(self.zpos * (self.zsize - 1))

        return self.zslice

    def yslice_from_ypos(self):
        self.yslice = int(self.ypos * (self.ysize - 1))

        return self.yslice

    def xslice_from_xpos(self):
        self.xslice = int(self.xpos * (self.xsize - 1))

        return self.xslice


class Pipeline:
    def __init__(self):
        self.figsize = 8  # Used for matplotlig figure size
        self.display_saturation_percentile = 99
        self.verbose = False

    def process_input(self, *args):
        session = args[0]
        session.data_path = args[1]

        # Get info from Tiff file
        session.get_data_info()
        session.create_summary_dict()

        # Memory map data as a virtual stack
        session.vol = DataLoader().load_tiff(session.data_path)

        if self.verbose:
            print(ouf.br(3, return_str=True) + session.summary_str())

        return session

    def show_summary_str(self, session):
        session.create_summary_dict()
        return session.summary_str()

    def create_zslice_fig(self, session):
        slice_fig = self.create_slice_fig("z", session)

        return slice_fig

    def create_yslice_fig(self, session):
        slice_fig = self.create_slice_fig("y", session)

        return slice_fig

    def create_xslice_fig(self, session):
        slice_fig = self.create_slice_fig("x", session)

        return slice_fig

    def create_slice_fig(self, axis, session):
        plt.close()
        vol = session.vol

        zslice = session.zslice_from_zpos()
        yslice = session.yslice_from_ypos()
        xslice = session.xslice_from_xpos()

        # Check if we something to use as vmin and vmax
        if session.min_percentile and session.max_percentile:
            vmin = session.min_percentile
            vmax = session.max_percentile
        else:
            vmin = None
            vmax = None

        if axis == "z":
            slice_fig = self._zslice_fig(vol, zslice, vmin=vmin, vmax=vmax)
        if axis == "y":
            slice_fig = self._yslice_fig(vol, yslice, vmin=vmin, vmax=vmax)
        if axis == "x":
            slice_fig = self._xslice_fig(vol, xslice, vmin=vmin, vmax=vmax)

        return slice_fig

    def _zslice_fig(self, vol, slice, **kwargs):
        fig = self.create_img_fig(vol[slice, :, :], **kwargs)

        return fig

    def _yslice_fig(self, vol, slice, **kwargs):
        fig = self.create_img_fig(vol[:, slice, :], **kwargs)

        return fig

    def _xslice_fig(self, vol, slice, **kwargs):
        fig = self.create_img_fig(vol[:, :, slice], **kwargs)

        return fig

    def create_img_fig(self, img, **kwargs):
        fig, ax = plt.subplots(figsize=(self.figsize, self.figsize))

        ax.imshow(img, interpolation="nearest", **kwargs)

        # Adjustments
        ax.axis("off")
        fig.subplots_adjust(left=0, right=1, bottom=0, top=1)

        return fig

    def create_projections_figs(self, session):
        vol = session.vol

        # Run projections
        max_projection, min_projection = self.get_projections(vol, session)

        # Generate figures
        max_projection_fig = self.create_img_fig(
            max_projection,
            vmin=session.min_percentile,
            vmax=session.max_percentile,
        )
        min_projection_fig = self.create_img_fig(
            min_projection,
            vmin=session.min_percentile,
            vmax=session.max_percentile,
        )
        return session, max_projection_fig, min_projection_fig

    def get_projections(self, vol, session):
        # Create arrays for iteration
        max_projection = np.zeros(np.shape(vol[0]))
        min_projection = np.ones(np.shape(vol[0])) * float("inf")
        intensity_sum = 0

        # Iterate over slices
        for zslice in vol:
            max_projection = np.maximum(max_projection, zslice)
            min_projection = np.minimum(min_projection, zslice)
            intensity_sum += np.sum(zslice)

        session.min_value = np.min(min_projection)
        session.min_percentile = np.percentile(
            min_projection, 100 - self.display_saturation_percentile
        )
        session.max_value = np.max(max_projection)
        session.max_percentile = np.percentile(
            max_projection, self.display_saturation_percentile
        )

        session.intensity_sum = intensity_sum

        nvoxels = session.zsize * session.ysize * session.xsize
        session.mean_intensity = intensity_sum / nvoxels
        return max_projection, min_projection

    def plot_vol_histogram(self, session):
        vol_hist, bin_edges = self.vol_histogram(
            session.vol, session.nbins, session.min_value, session.max_value
        )

        fig, ax = plt.subplots(figsize=(6, 4))

        ax.bar(
            bin_edges[:-1], vol_hist, width=np.diff(bin_edges), ec="white", align="edge"
        )

        # Adjustments
        ax.spines["right"].set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["left"].set_visible(True)
        ax.spines["bottom"].set_visible(True)
        ax.set_yscale("log")

        # fig.subplots_adjust(left=0, right=1, bottom=0, top=1)

        return fig

    def vol_histogram(self, vol, nbins, min_value, max_value):
        # Start histogram
        vol_hist = np.zeros(nbins)

        # Iterate over slices
        for zslice in vol:
            hist, bin_edges = np.histogram(
                zslice, bins=nbins, range=(min_value, max_value)
            )
            vol_hist += hist

        return vol_hist, bin_edges


if __name__ == "__main__":
    app = Interface()
    app.show_header = True
    app.launch(server_name="0.0.0.0", show_error=True)



class DataSaver:
    def __init__(self):
        self.verbose = False

    def save(self, path, data):
        raise NotImplementedError("Save is not implemented yet")


def save(path, data):
    return DataSaver().save(path, data)



import socket
import hashlib


def mock_plot():
    import matplotlib.pyplot as plt
    from matplotlib.figure import Figure
    import numpy as np
    import matplotlib

    matplotlib.use("Agg")

    fig = plt.figure(figsize=(5, 4))
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
    xx = np.arange(0, 2 * np.pi, 0.01)
    ax.plot(xx, np.sin(xx))

    return fig


def mock_write_file(path):
    f = open(path, "w")
    f.write("File created by apptools")
    f.close()


def get_local_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(("192.255.255.255", 1))
        IP = s.getsockname()[0]
    except:
        IP = "127.0.0.1"
    finally:
        s.close()
    return IP


def port_from_str(s):
    return int(hashlib.sha1(s.encode("utf-8")).hexdigest(), 16) % (10**4)


def gradio_header(title, port):
    import outputformat as ouf

    ouf.br(2)
    details = [
        f'{ouf.c(title, color="rainbow", cmap="cool", bold=True, return_str=True)}',
        f"Using port {port}",
        f"Running at {get_local_ip()}",
    ]
    ouf.showlist(details, style="box", title="Starting gradio server")


def sizeof(num, suffix="B"):
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f} {unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f} Y{suffix}"
